%--------------------------------------------------------------------------
function probSame = Summary_ChangeDetection(firstDisplay, secondDisplay, numToRemember)
    
    % Get likelihood of various gist memory params based on the first display
    gist.horiz = -1.5:0.5:1.5;
    gist.vert  = -1.5:0.5:1.5;
    gist.like = Exact_MRF_Parameters(firstDisplay, gist.horiz, gist.vert);
    
    % Setup output matrices
    probSamePosterior = zeros(length(gist.horiz), length(gist.vert), nchoosek(25,numToRemember));
    probEncode = zeros(length(gist.horiz), length(gist.vert), nchoosek(25,numToRemember));
    
    % For each possible gist we could encode....
    for h=1:length(gist.horiz)
        for v=1:length(gist.vert)
            gistParams = [gist.horiz(h) gist.vert(v)];
            
            % Find the numToRemember most outlier items and how much of an outlier they are
            informativeItems = CalcMostInformativeItems(...
                gistParams, firstDisplay, numToRemember);
            
            % For each set of items we could encode from the display...
            for w=1:size(informativeItems.possibleRemember,1)
                
                % Get posterior given those items
                itemMemory = informativeItems.possibleRemember(w,:);
                probSame = PosteriorOnChangedDisplay(gistParams, secondDisplay, ...
                        firstDisplay, itemMemory, numToRemember);

                % Store
                probEncode(h,v,w) = informativeItems.encodeThis(w) * gist.like(h,v);
                probSamePosterior(h,v,w) = probSame;
            end
        end
    end
    
    % Marginalize over the gist parameters and item memory parameters
    probEncode = probEncode ./ sum(probEncode(:));
    probSame = sum(probSamePosterior(:) .* probEncode(:));
end

%--------------------------------------------------------------------------
function items = CalcMostInformativeItems(gistParams, ...
        firstDisplay, numToRemember)
    
    % Figure out what the most informative items are
    outlierMap = CondProbEachItem(gistParams, firstDisplay);
    
    % Calculate S, the set of items we might choose, as well as their
    % probability of being chosen (based on how much of an outlier the
    % chosen and not chosen items are)
    
    % Calculate the numToRemember most outlier items
    [~,b] = sort(outlierMap(:), 'descend');
    which = b(1:numToRemember);
    outlierMap = outlierMap(which);
    
    % Consider all possible sets of these items we could remember (of size
    % 0, 1 ... numToRemember)
    items.whichToRemember = logical(fullfact(zeros(1,numToRemember)+2) - 1);
    for i=1:size(items.whichToRemember,1)
        % Probability of encoding this set
		items.encodeThis(i) = prod(outlierMap(items.whichToRemember(i,:))) ...
            * prod(1 - outlierMap(~items.whichToRemember(i,:)));
        
        % Pad row with -1's 
        row = zeros(1, numToRemember) - 1;
        chosenItems = sort(which(items.whichToRemember(i,:)));
        row(1:length(chosenItems)) = chosenItems;
		items.possibleRemember(i,:) = row;
    end
    items.encodeThis = items.encodeThis ./ sum(items.encodeThis(:));
end

%--------------------------------------------------------------------------
function probSame = PosteriorOnChangedDisplay(gistParams, ...
        displayTwo, displayOne, lockedItems, numToRemember)
    
    % Priors on the likelihood of a change
    priorNoChange = 0.5;
    priorChange = 0.5 * 1/numel(displayTwo);
    
    % Calculate how likely any given location is to have changed.
    probOrigDisplay = ExactLike(gistParams, displayTwo);
    newInforItems = CalcMostInformativeItems(gistParams, displayTwo, numToRemember);
    probItems = ChanceEncodingThese(newInforItems, lockedItems);
    probNoChange = priorNoChange * probOrigDisplay * probItems;
    
    for n=1:length(displayTwo(:))
        % For each item, we are sure whether it changed or not if
        % we remembered it from display one
        if any(lockedItems == n)
            if displayOne(n) == displayTwo(n)
                probChange(n) = 0;
            else
                probChange(n) = Inf;
            end
        else
            % If we didn't remember it, we try to infer whether or
            % not it changed based on the gist and the other items
            cur = displayTwo;
            cur(n) = 1-cur(n);
            
            % Calculate the probability of the original display that would
            % have existed if this had been the change, given the gist we
            % encoded
            probChangedDisplay = ExactLike(gistParams, cur);
            
            % Figure out how likely we are to have chosen the items we did
            % under this hypothesis about what the original display looked
            % like
            newInforItems = CalcMostInformativeItems(gistParams, ...
                cur, numToRemember);
            probItems = ChanceEncodingThese(newInforItems, lockedItems);
            
            % Calculate overall chance of this change/this old display
            probChange(n) = priorChange * probChangedDisplay * probItems;
        end
    end
    probSame = probNoChange ./ (probNoChange + sum(probChange));
end

%--------------------------------------------------------------------------
function probItems = ChanceEncodingThese(itemsNew, otherItemList)
    % What is the probability we would have encoded the set of items we have
    % in memory (otherItemList), given the most informative items in the
    % display we are imagining might have been the first display (itemsNew)
    probItems = itemsNew.encodeThis(...
        ismember(itemsNew.possibleRemember, otherItemList, 'rows') );
    if isempty(probItems)
        probItems = 0;
    end
end

%--------------------------------------------------------------------------
function exactLike = Exact_MRF_Parameters(cur, horiz, vert)
    % Calculate exact likelihood for grid of MRF parameters
    exactLike  = zeros(length(horiz),length(vert));
    for i=1:length(horiz)
        for j=1:length(vert)
            exactLike(i,j) = ExactLike([horiz(i) vert(j)], cur);
        end
    end
    exactLike = exactLike ./ sum(exactLike(:));
end

%--------------------------------------------------------------------------
function probMat = CondProbEachItem(params, cur)
    % Calculate how much of an outlier each item is
    probMat = zeros(size(cur));
    start = ExactLike(params, cur);
    for n=1:length(cur(:))
        flippedDisplay = cur;
        flippedDisplay(n) = 1-cur(n);
        probMat(n) = start / (start+ExactLike(params, flippedDisplay));
    end
    probMat = 1 - probMat;
end

%--------------------------------------------------------------------------
function like = ExactLike(params, disp)
    % Calculate the exact likelihood of this display
    persistent hPairs; persistent vPairs; persistent z;
    persistent horiz; persistent vert;
    if isempty(hPairs)
        % For speed: Cache the item indexes that are adjacent
        % horizontally and vertically so we don't have to recalculate them
        % everytime
        nn = (1:numel(disp))';
        nnD=nn+1;	nnR=nn+size(disp,1);
        hPairs = [nn nnR];
        hPairs = hPairs(nnR>=1&nnR<=numel(disp),:);
        vPairs = [nn nnD];
        cols = ceil(vPairs./size(disp,1));
        vPairs = vPairs(nnD>=1&nnD<=numel(disp)&(cols(:,1)==cols(:,2)),:);
        
        % Load the precalculated normalizing constants for each set of
        % parameters - this loads horiz, vert and z
        load Summary_z/z5x5.mat; % Get Z
    end
    
    % Find the correct normalizing constant
    which = [find(horiz==params(1)) find(vert==params(2))];
    curZ = z(which(1),which(2));
    
    % Calculate likelihood
    valsH = disp(hPairs);
    valsV = disp(vPairs);
    hSum = sum(valsH(:,1)==valsH(:,2))*1 + sum(valsH(:,1)~=valsH(:,2))*-1;
    vSum = sum(valsV(:,1)==valsV(:,2))*1 + sum(valsV(:,1)~=valsV(:,2))*-1;
    like = exp(-params(1)*hSum + -params(2)*vSum) ./ curZ;
end


