%--------------------------------------------------------------------------
function probSame = Summary_ChangeDetection(curD, curD2, numToRemember)
    
    % Get likelihood of various gist memory params based on the first display
    gist.horiz = -5:1:5;
    gist.vert  = -5:1:5;
    gist.like = Exact_MRF_Parameters(curD, gist.horiz, gist.vert);
    
    % Setup output matrices
    probSamePosterior = zeros(length(gist.horiz), length(gist.vert), nchoosek(8,numToRemember));
    probEncode = zeros(length(gist.horiz), length(gist.vert), nchoosek(8,numToRemember));
    
    % For each possible gist we could encode....
    for h=1:length(gist.horiz)
        for v=1:length(gist.vert)
            gistParams = [gist.horiz(h) gist.vert(v)];
            
            % Find the numToRemember most outlier items and how much of an outlier they are
            informativeItems = CalcMostInformativeItems(...
                gistParams, curD, numToRemember);
            
            % For each set of items we could encode from the display...
            for w=1:size(informativeItems.possibleRemember,1)
                
                % Calculate the posterior of the second display given possible changes
                % to the first display and the gist and item memory components
                itemMemory = informativeItems.possibleRemember(w,:);
								probSame = PosteriorOnChangedDisplay(gistParams, curD2, ...
                    curD, itemMemory, numToRemember);
                
                % Store
                probEncode(h,v,w) = informativeItems.encodeThis(w) * gist.like(h,v);
                probSamePosterior(h,v,w) = probSame;
            end
        end
    end
    
    % Marginalize over the gist parameters and item memory parameters
    probSame = sum(probSamePosterior(:) .* probEncode(:));
end

%--------------------------------------------------------------------------
function items = CalcMostInformativeItems(gistParams, ...
        curD, numToRemember)
    
    % Figure out what the most informative items are
    [~,map] = ExactLike(gistParams, curD);
    outlierMap = 1-map;
    
    % Calculate S, the set of items we might choose, as well as their
    % probability of being chosen (based on how much of an outlier the
    % chosen and not chosen items are)
    items.possibleRemember = [];
    items.encodeThis = [];
    
    % For each # of items we could remember, K, K-1, ... 0
    for r=numToRemember:-1:0
        % Get a list of all the possible ways of choosing r items from the
        % set of items, plus the probability of choosing them (based on
        % outlier map)
        [posRemember, encodeThis] = ...
            AllPossibleItemsAndProbs(1:numel(outlierMap), r, outlierMap(:));
        
        % Set any unused columns to negative 1
        posRememberPadded = zeros(size(posRemember,1), numToRemember) - 1;
        
        % Fill in the actual set of items
        posRememberPadded(:,1:size(posRemember,2)) = posRemember;
        
        % Combine these into the larger list
        items.possibleRemember = [items.possibleRemember; posRememberPadded];
        items.encodeThis = [items.encodeThis; encodeThis];
    end
    items.encodeThis = items.encodeThis ./ numToRemember; % Normalize
end

%--------------------------------------------------------------------------
function [posRemember, encodeThis] = AllPossibleItemsAndProbs(items, numToRemember, outlierMap)
    % NOTE: A mex version of this function is available in MexFiles/
    % Just replace this function with that one to gain a considerable speed
    % improvement
    
    % All possible ways of choose numToRemember items from the display
    posRemember = nchoosek(items, numToRemember);
    
    % For each set of items, calculate the probability of choosing it based on
    % outlierMap
    encodeThis = zeros(size(posRemember,1), 1);
    for i=1:size(posRemember,1)
        doRemember = 1-outlierMap(:);
        doRemember(posRemember(i,:)) = 1-doRemember(posRemember(i,:));
        encodeThis(i) = prod(doRemember);
    end
    encodeThis = encodeThis ./ sum(encodeThis(:));
end

%--------------------------------------------------------------------------
function probSame = PosteriorOnChangedDisplay(gistParams, ...
        curD2, curD, lockedItems, numToRemember)
    
    % Priors on the likelihood of a change
    priorNoChange = 0.5;
    
    % There are numel(..) items, and since each can be 7 colors, there are
    % 6 ways each of them could change
    priorChange = 0.5 * 1/(numel(curD2.colorItems)*(7-1));
    
    % Calculate how likely any given location is to have changed.
    probOrigDisplay = ExactLike(gistParams, curD2);
    newInforItems = CalcMostInformativeItems(gistParams, curD2, numToRemember);
    probItems = ChanceEncodingThese(newInforItems, lockedItems);
    probNoChange = priorNoChange * probOrigDisplay * probItems;
    
    probChange = zeros(1, length(curD2.colorItems));
    for n=1:length(curD2.colorItems)
        % For each item, we are sure whether it changed or not if
        % we remembered it from display one
        if any(lockedItems == n)
            if curD.colorItems(n) == curD2.colorItems(n)
                probChange(n) = 0;
            else
                probChange(n) = Inf;
            end
        else
            % If we didn't remember it, we try to infer whether or
            % not it changed based on the gist and the other items
            probChange(n) = 0;
            for c=1:7 % Number of colors
                cur = curD2;
                if curD2.colorItems(n)==c
                    continue;
                end
                cur.colorItems(n) = c;
                
                % Calculate the probability of the original display that would
                % have existed if this had been the change, given the gist we
                % encoded
                probChangedDisplay = ExactLike(gistParams, cur);
                
                % Figure out how likely we are to have chosen the items we did
                % under this hypothesis about what the original display looked
                % like
                [newInforItems] = CalcMostInformativeItems(gistParams, ...
                    cur, numToRemember);
                probItems = ChanceEncodingThese(newInforItems, lockedItems);
                
                % Calculate overall chance of this change/this old display
                probChange(n) = probChange(n) + (priorChange * probChangedDisplay * probItems);
            end
        end
    end
    probSame = probNoChange ./ (probNoChange + sum(probChange));
end

%--------------------------------------------------------------------------
function probItems = ChanceEncodingThese(itemsNew, otherItemList)
    % What is the probability we would have encoded the set of items we have
    % in memory (otherItemList), given the most informative items in the
    % display we are imagining might have been the first display (itemsNew)
    probItems = itemsNew.encodeThis(...
        ismember(itemsNew.possibleRemember, otherItemList, 'rows') );
end

%--------------------------------------------------------------------------
function exactLike = Exact_MRF_Parameters(cur, horiz, vert)
    % Calculate exact likelihood for grid of MRF parameters
    exactLike  = zeros(length(horiz),length(vert));
    for i=1:length(horiz)
        for j=1:length(vert)
            exactLike(i,j) = ExactLike([horiz(i) vert(j)], cur);
        end
    end
    exactLike = exactLike ./ sum(exactLike(:));
end

%--------------------------------------------------------------------------
function [l, map] = ExactLike(params, curD)
    % NOTE: A mex version of this function is available in MexFiles/
    % Just replace this function with that one to gain a considerable speed
    % improvement
    
    % This function uses the pseudo-likelihood to calculate the probability of
    % any given display under the parameters
    persistent hPairs; persistent vPairs;
    if isempty(hPairs)
        % For speed: Cache the item indexes that are adjacent
        % horizontally and vertically so we don't have to recalculate them
        % everytime
        dispSize = [4 5];
        nn = (1:prod(dispSize))';
        nnD=nn+1;	nnR=nn+dispSize(1);
        hPairs = [nn nnR];
        hPairs = hPairs(nnR>=1&nnR<=prod(dispSize),:);
        vPairs = [nn nnD];
        cols = ceil(vPairs./dispSize(1));
        vPairs = vPairs(nnD>=1&nnD<=prod(dispSize)&(cols(:,1)==cols(:,2)),:);
    end
    
    % Initialize variables
    pdiffHOthers = zeros(length(curD.location), 7);
    pdiffVOthers = zeros(length(curD.location), 7);
    map = zeros(1, length(curD.location));
    
    l = 1;
    % For each item....
    for i=1:length(curD.location)
        
        % Consider each color it could be...
        for j=1:7
            cColors = curD.colorItems;
            cColors(i) = j;
            
            % Make a 4x5 display with the colors in the right
            % locations
            curColors = nan(1, 4*5);
            curColors(curD.location) = cColors;
            
            % And consider the likelihood of that display....
            hPairsNew = curColors(hPairs);
            vPairsNew = curColors(vPairs);
            
            pdiffHOthers(i,j) = sum(hPairsNew(:,1)==hPairsNew(:,2));
            pdiffVOthers(i,j) = sum(vPairsNew(:,1)==vPairsNew(:,2));
        end
        
        % Divide the likelihood of the current display...
        probThis = exp(params(1)*pdiffHOthers(i,curD.colorItems(i)) + params(2)*pdiffVOthers(i,curD.colorItems(i)));
        
        % By the sum of the likelihood's of all possible versions of this
        % item
        probAll  = sum(exp(params(1).*pdiffHOthers(i,:) + params(2).*pdiffVOthers(i,:)));
        
        % Store this as the conditional probability of this item (which we
        % return, since we need it for picking outliers anyway)
        map(i) = (probThis./probAll);
        
        % And accumulate those conditional probabilities to get the
        % pseudolikelihood of the entire display
        l = l * (probThis./probAll);
    end
end





